# Copyright (c) 2024 iiPython

# Modules
import typing
from time import time
from json import JSONDecodeError
from secrets import token_urlsafe

from pydantic import BaseModel, Field

from requests import Session
from fastapi import FastAPI, WebSocket
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from starlette.websockets import WebSocketDisconnect, WebSocketState

from nightwatch import __version__
from nightwatch.logging import log
from nightwatch.config import fetch_config

# Load config data
config = fetch_config("rics")

# Initialization
app = FastAPI(openapi_url = None)
app.add_middleware(CORSMiddleware, allow_origins = ["*"], allow_methods = ["*"])

session = Session()

# Check for updates
app.state.latest_update = None
if config["enable_update_checking"] is not False:
    latest = session.get("https://api.github.com/repos/iiPythonx/nightwatch/releases/latest").json()

    def version(v: str) -> tuple:
        return tuple(map(int, v.split(".")))

    if version(latest["name"][1:]) > version(__version__):
        log.info("update", f"Nightwatch {latest['name']} is now available, upgrading is recommended.")
        log.info("update", f"See the changelog at {latest['html_url']}.")
        app.state.latest_update = latest["name"][1:]

# Scaffold the application
app.state.clients = {}
app.state.pending_clients = {}
app.state.message_log = []

async def broadcast(payload: dict) -> None:
    payload["data"]["time"] = round(time())
    if payload["type"] == "message":
        if "user" not in payload["data"]:
            payload["data"]["user"] = {"name": "Nightwatch", "hex": "555753", "admin": False, "bot": True}

        app.state.message_log = app.state.message_log[-24:] + [payload["data"]]

    for client in app.state.clients.values():
        await client.send(payload)

app.state.broadcast = broadcast

# Setup routing
class Client:
    def __init__(self, websocket: WebSocket, user_data: dict[str, typing.Any]) -> None:
        self.websocket = websocket
        self.username, self.hex_code = user_data["username"], user_data["hex"]

        # Attributes
        self.admin, self.bot = False, user_data["bot"]

        # Attach to client list
        self._callback = None
        app.state.clients[self.username] = self

    def serialize(self) -> dict[str, str | bool]:
        return {"name": self.username, "hex": self.hex_code, "admin": self.admin, "bot": self.bot}

    def cleanup(self) -> None:
        del app.state.clients[self.username]
        del self  # Not sure if this helps, in case Python doesn't GC

    async def send(self, payload: dict) -> None:
        if self.websocket.application_state != WebSocketState.CONNECTED:
            return

        try:
            if self._callback is not None and payload["type"] != "message":
                payload["data"] = payload.get("data", {}) | {"callback": self._callback}
                self._callback = None

            await self.websocket.send_json(payload)

        except WebSocketDisconnect:
            pass

    async def receive(self) -> typing.Any:
        try:
            data = await self.websocket.receive_json()

            # Handle callback
            callback = data.get("data", {}).get("callback")
            if isinstance(callback, str):
                self._callback = callback

            return data

        except KeyError:
            await self.websocket.close(1002, "Nightwatch uses text frames, binary frames are rejected.")

        except JSONDecodeError:
            await self.websocket.close(1002, "Nightwatch requires all packets to be sent using JSON, non-JSON is rejected.")

        except WebSocketDisconnect:
            pass

        except Exception:
            await self.websocket.close(1002, "Some data parsing issue occured, check your payloads.")

        return None

class ClientJoinModel(BaseModel):
    username: str = Field(..., min_length = 3, max_length = 30)
    hex: str = Field(..., min_length = 6, max_length = 6, pattern = "^[0-9A-Fa-f]{6}$")
    bot: bool = False

@app.post("/api/join")
async def route_index(client: ClientJoinModel) -> JSONResponse:
    if client.username in app.state.clients:
        return JSONResponse({
            "code": 400,
            "message": "Requested username is in use on this server."
        }, status_code = 400)

    if client.username.strip() != client.username:
        return JSONResponse({
            "code": 400,
            "message": "Requested username has whitespace that should be removed prior to joining."
        }, status_code = 400)

    if client.username.lower() in ["nightwatch", "admin", "moderator"]:
        return JSONResponse({
            "code": 400,
            "message": "Requested username is restricted for use."
        }, status_code = 400)

    client_token = token_urlsafe()
    app.state.pending_clients[client_token] = client.model_dump()
    return JSONResponse({
        "code": 200,
        "authorization": client_token
    })

@app.websocket("/api/ws")
async def connect_endpoint(
    authorization: str,
    websocket: WebSocket
) -> None:
    if authorization not in app.state.pending_clients:
        return await websocket.close(1008)

    user_data = app.state.pending_clients[authorization]
    del app.state.pending_clients[authorization]

    await websocket.accept()

    # Initialize client
    client = Client(websocket, user_data)

    # Get the client up to speed
    await client.send({"type": "rics-info", "data": {
        "name": config["name"] or "Nightwatch Server",
        "message-log": app.state.message_log,
        "user-list": [client.serialize() for client in app.state.clients.values()]
    }})

    # Broadcast join
    await app.state.broadcast({"type": "join", "data": {"user": client.serialize()}})
    await app.state.broadcast({"type": "message", "data": {"message": f"{client.username} has joined the server."}})

    # Handle loop
    while websocket.application_state == WebSocketState.CONNECTED:
        match await client.receive():
            case {"type": "message", "data": {"message": message}}:
                if not message.strip():
                    await client.send({"type": "problem", "data": {"message": "You cannot send a blank message."}})
                    continue

                await app.state.broadcast({"type": "message", "data": {"user": client.serialize(), "message": message}})
                if client._callback is not None:
                    await client.send({"type": "response"})

            case {"type": "user-list", "data": _}:
                await client.send({"type": "response", "data": {
                    "user-list": [client.serialize() for client in app.state.clients.values()]
                }})

            case None:
                break

            case _:
                await client.send({"type": "problem", "data": {"message": "Invalid payload received."}})

    client.cleanup()
    await app.state.broadcast({"type": "leave", "data": {"user": client.serialize()}})
    await app.state.broadcast({"type": "message", "data": {"message": f"{client.username} has left the server."}})

@app.get("/api/version")
async def route_version() -> JSONResponse:
    return JSONResponse({"code": 200, "data": {"version": __version__, "latest": app.state.latest_update}})

# Load additional routes
from nightwatch.rics.routing import (  # noqa: E402
    files,          # noqa: F401
    image_forward   # noqa: F401
)



# Copyright 2022 iiPython

# Modules
import string
import msvcrt
from typing import Union

from . import keys

# String list
string_check = string.ascii_letters + string.digits + string.punctuation

# Main function
def readchar() -> Union[str, int]:

    # Grab a character
    ch = msvcrt.getch()
    ch_dc = ch.decode("mbcs")
    if ch_dc in string_check:
        return ch_dc

    ch = ord(ch)
    if ch == 0 or ch == 224:
        ch = ord(msvcrt.getch())

    # Character map
    if ch == keys.SPACE:
        return " "

    return ch



# Copyright (c) 2024 iiPython

# Modules
import os
import re
import typing
import requests
from threading import Thread
from argparse import ArgumentParser

import urwid
import websockets
from websockets.sync.client import connect

from nightwatch import __version__, HEX_COLOR_REGEX

from . import config
from .extra.ui import NightwatchUI
from .extra.select import menu
from .extra.wswrap import ORJSONWebSocket

# Initialization
if os.name == "nt":
    urwid.set_encoding("utf-8")

# Connection handler
def connect_loop(host: str, port: int, username: str) -> None:
    protocol, url = "s" if port == 443 else "", f"{host}:{port}"

    # Perform authentication
    resp = requests.post(f"http{protocol}://{url}/api/join", json = {"username": username, "hex": config["client.color"]}).json()
    if resp.get("code") != 200:
        exit(f"\nCould not connect to {url}. Additional details:\n{resp}")

    destination = f"ws{protocol}://{url}/api/ws?authorization={resp['authorization']}"
    try:
        with connect(destination) as ws:
            ws = ORJSONWebSocket(ws)

            # Handle fetching server information
            response = ws.recv()

            # Create UI
            ui = NightwatchUI(ws)
            loop = urwid.MainLoop(ui.frame, [
                ("yellow", "yellow", ""),
                ("gray", "dark gray", "", "", "#555753", ""),
                ("green", "dark green", "")
            ])
            loop.screen.set_terminal_properties(2 ** 24)  # Activate 24-bit color mode

            # Individual components
            loop.screen.register_palette_entry("time", "dark green", "", foreground_high = config["colors.time"] or "#00FF00")
            loop.screen.register_palette_entry("sep", "dark gray", "", foreground_high = config["colors.sep"] or "#555753")

            # Handle messages
            def message_loop(ws: ORJSONWebSocket, ui: NightwatchUI) -> None:
                try:
                    while ws.ws:
                        ui.on_message(ws.recv())

                except websockets.exceptions.ConnectionClosed:
                    pass

            Thread(target = message_loop, args = [ws, ui]).start()

            # Start mainloop
            ui.on_ready(loop, response["data"])
            loop.run()

    except websockets.exceptions.InvalidURI:
        exit(f"\nCould not connect to {destination} due to an HTTP redirect.\nPlease ensure you entered the correct address.")

    except OSError:
        exit(f"\nCould not connect to {destination} due to an OSError.\nThis is more then likely because the server is not running.")

# Entrypoint
def start_client(
    address: typing.Optional[str] = None,
    username: typing.Optional[str] = None
):
    username = username or config["client.username"]

    # Start main UI
    print(f"\033[H\033[2J✨ Nightwatch | v{__version__}\n")
    if username is None:
        print("Hello! It seems that this is your first time using Nightwatch.")
        print("Before you can connect to a server, please set your desired username.\n")

        username = input("Username: ")
        config.set("client.username", username)
        print("\033[4A\033[0J", end = "")  # Reset back up to the Nightwatch label

    # Handle color setup
    color = config["client.color"] or ""
    if not re.match(HEX_COLOR_REGEX, color):
        while True:
            print("For fun, you can select a color for your username.")
            print("Please enter the HEX code (6 long) you would like to have as your color.")
            color = (input("> #") or "ffffff").lstrip("#")

            # Validate their color choice
            if re.match(HEX_COLOR_REGEX, color):
                break

            print("\033[3A\033[0J", end = "")

        print("\033[3A\033[0J", end = "")
        config.set("client.color", color)

    # Handle server address
    if address is None:
        servers = config["client.servers"]
        if servers is None:
            servers = ["nightwatch.iipython.dev"]
            config.set("client.servers", servers)

        print(f"Hello, {username}. Please select a Nightwatch server to connect to:")
        address = menu.show(servers)
        print()

    print(f"Establishing connection to {address} ...")

    # Parse said address
    if ":" not in address:
        host, port = address, 443

    else:
        host, port = address.split(":")

    # Connect to server
    try:
        connect_loop(host, int(port), username)

    except KeyboardInterrupt:
        print("\033[5A\033[0J", end = "")  # Reset back up to the Nightwatch label
        print(f"Goodbye, {username}.")

# Initialization
def main() -> None:
    ap = ArgumentParser(
        prog = "nightwatch",
        description = "The chatting application to end all chatting applications.\nhttps://github.com/iiPythonx/nightwatch",
        epilog = "Copyright (c) 2024 iiPython"
    )
    ap.add_argument("-a", "--address", help = "the nightwatch server to connect to")
    ap.add_argument("-u", "--username", help = "the username to use")
    ap.add_argument("-r", "--reset", action = "store_true", help = "reset the configuration file")

    # Launch client
    args = ap.parse_args()
    if args.reset:
        fetch_config("config").reset()

    start_client(args.address, args.username)



import os
from . import keys

if os.name == "nt":
    import msvcrt
    from .read import readchar

elif "nix" in os.name or os.name == "posix":
    from .read_linux import readchar

else:
    raise OSError("unsupported OS '{}'".format(os.name))



# Copyright (c) 2024 iiPython

# Modules
from nightwatch.config import config_path

# Initialization
class NightwatchLogger:
    def __init__(self) -> None:
        self._color_map = {
            "info": "34", "warn": "33", "error": "31", "critical": "31"
        }
        self._log_file = config_path / "nightwatch.log"

    def log(self, level: str, component: str, message: str) -> None:
        print(f"\033[{self._color_map[level]}m⚡︎ {level.upper()} ({component}): {message}\033[0m")
        with self._log_file.open("a") as log_file:
            log_file.write(f"⚡︎ {level.upper()} ({component}): {message}\n")

    # Submethods
    def info(self, *args) -> None:
        self.log("info", *args)

    def warn(self, *args) -> None:
        self.log("warn", *args)

    def error(self, *args) -> None:
        self.log("error", *args)

    def critical(self, *args) -> None:
        self.log("critical", *args)
        exit(1)

log = NightwatchLogger()



# Copyright (c) 2024 iiPython

# Modules
from typing import Callable

from nightwatch import __version__
from nightwatch.client import config

# Main class
class BaseCommand():
    def __init__(self, name: str, ui, add_message: Callable) -> None:
        self.name, self.ui = name, ui
        self.add_message = add_message

    def print(self, message: str) -> None:
        self.add_message(self.name.title(), message)

# Commands
class ShrugCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("shrug", *args)

    def on_execute(self, args: list[str]) -> str:
        return r"¯\_(ツ)_/¯"

class ConfigCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("config", *args)

    def on_execute(self, args: list[str]) -> None:
        if not args:
            for line in [
                "Nightwatch client configuration",
                "Usage: /config <key> <value>",
                "",
                "Example usage:",
                "  /config colors.time yellow",
                "  /config prompt \">> \"",
                "  /config client.time_format 24h",
                "",
                "Some changes will only apply after Nightwatch restarts."
            ]:
                self.print(line)

            return

        elif len(args) < 2:
            return self.print(f"Missing the value to assign to '{args[0]}'.")

        config.set(args[0], args[1])
        self.print(f"{args[0]} has been set to \"{args[1]}\".")

class HelpCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("help", *args)

    def on_execute(self, args: list[str]) -> None:
        self.print(f"✨ Nightwatch v{__version__}")
        self.print("Available commands:")
        for command in self.ui.commands:
            self.print(f"  /{command}")

class MembersCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("members", *args)

    def on_execute(self, args: list[str]) -> None:
        def members_callback(response: dict):
            self.print(", ".join(user['name'] for user in response["data"]["user-list"]))

        self.ui.websocket.callback({"type": "user-list"}, members_callback)

class AdminsCommand(BaseCommand):
    def __init__(self, *args) -> None:
        super().__init__("admins", *args)

    def on_execute(self, args: list[str]) -> None:
        def admins_callback(response: dict):
            if not response["data"]["admin-list"]:
                return self.print("There are no connected admins.")

            self.print(f"Connected admins: {', '.join(admin['name'] for admin in response['data']['admin-list'])}")

        self.ui.websocket.callback({"type": "admin-list"}, admins_callback)

# class AdminCommand(BaseCommand):
#     def __init__(self, *args) -> None:
#         self.admin = False
#         super().__init__("admin", *args)

#     def on_execute(self, args: list[str]) -> None:
#         match args:
#             case [] if not self.admin:
#                 self.ui.websocket.send({"type": "admin"})
#                 self.print("Run /admin <code> with the admin code in your server console.")

#             case [] | ["help"]:
#                 self.print("Available commands:")
#                 if not self.admin:
#                     self.print("  /admin <admin code>")

#                 self.print("  /admin ban <username>")
#                 self.print("  /admin unban <username>")
#                 self.print("  /admin ip <username>")
#                 self.print("  /admin banlist")
#                 self.print("  /admin say <message>")

#             case [code] if not self.admin:
#                 def on_admin_response(response: dict):
#                     if response["data"]["success"] is False:
#                         return self.print("(fail) Invalid admin code specified.")

#                     self.print("(success) Privileges escalated.")
#                     self.admin = True

#                 self.ui.websocket.callback({"type": "admin", "data": {"code": code}}, on_admin_response)

#             case ["ban", username]:
#                 def on_ban_response(response: dict):
#                     if not response["data"]["success"]:
#                         return self.print(f"(fail) {response['data']['error']}")

#                     self.print(f"(success) {username} has been banned.")

#                 self.ui.websocket.callback({"type": "admin", "data": {"command": args}}, on_ban_response)

#             case ["unban", username]:
#                 def on_unban_response(response: dict):
#                     if not response["data"]["success"]:
#                         return self.print(f"(fail) {response['data']['error']}")

#                     self.print(f"(success) {username} has been unbanned.")

#                 self.ui.websocket.callback({"type": "admin", "data": {"command": args}}, on_unban_response)

#             case ["ip", username]:
#                 def on_ip_response(response: dict):
#                     if not response["data"]["success"]:
#                         return self.print(f"(fail) {response['data']['error']}")

#                     self.print(f"(success) {username}'s IP address is {response['data']['ip']}.")

#                 self.ui.websocket.callback({"type": "admin", "data": {"command": args}}, on_ip_response)

#             case ["banlist"]:
#                 def on_banlist_response(response: dict):
#                     if not response["data"]["banlist"]:
#                         return self.print("(fail) Nobody is banned on this server.")

#                     self.print("Current banlist:")
#                     self.print(f"{', '.join(f'{v} ({k})' for k, v in response['data']['banlist'].items())}")

#                 self.ui.websocket.callback({"type": "admin", "data": {"command": args}}, on_banlist_response)

#             case ["say", _]:
#                 self.ui.websocket.send({"type": "admin", "data": {"command": args}})

#             case _:
#                 self.print("Admin command not recognized, try /admin help.")

commands = [
    ShrugCommand,
    ConfigCommand,
    HelpCommand,
    MembersCommand,
    AdminsCommand,
    # AdminCommand
]



# Copyright 2022 iiPython

# Modules
import re
import os

# Color map
def to_ansi(code: int) -> str:
    return f"\033[{code}m"

colormap_ = {
    "black": 30, "red": 31, "green": 32,
    "yellow": 33, "blue": 34, "magenta": 35,
    "cyan": 36, "white": 37, "bright": 1,
    "dim": 2, "norm": 22, "reset": 39
}
colormap = {}
for color, code in colormap_.items():
    colormap[color] = to_ansi(code)
    if color not in ["bright", "dim", "norm"]:
        colormap["l" + color] = to_ansi(code + 60)
        colormap["bg" + color] = to_ansi(code + 10)
        colormap["bgl" + color] = to_ansi(code + 70)

# Initialization
_tagregex = re.compile(r"\[\/?\w*\]")
_clear_command = "cls" if os.name == "nt" else "clear"

# Terminal functions
def clear() -> None:
    """Clears the terminal screen"""
    os.system(_clear_command)

def color(text: str, dry: bool = False) -> str:
    """Colorizes text using bbcode-like markup

    Parameters:
        text (str): the text to colorize
        dry (bool): dry run, only strip tags

    Returns:
        text (str): the colorized text

    Example:
        .color("[red]Hello, red world![/red]")
        Returns:
            \033[31mHello, red world!\033[39m

        For more examples, see the Python rich library:
        https://github.com/willmcgugan/rich
    """
    text = str(text)

    # Process tags
    text += ("[reset]" if not text.endswith("[reset]") else "")
    tags = []
    for tag in _tagregex.findall(text):
        tag = tag.strip("[]")
        if tag == "/":
            isbg, tags = tags[-1][:2] == "bg", tags[:-1]
            text = text.replace(f"[/{tag[1:]}]", (colormap["bgreset"] if isbg else "") + (colormap[tags[-1] if tags else "reset"]) if not dry else "", 1)

        else:
            if tag not in colormap:
                continue

            text = text.replace(f"[{tag}]", colormap[tag] if not dry else "", 1)
            tags.append(tag)

    return text

def cprint(*args, **kwargs) -> None:
    """A wrapper for print() which calls .color() on each argument"""
    return print(*[color(a) for a in args], **kwargs)



# Copyright (c) 2024 iiPython

# Modules
import typing
import asyncio

import orjson
import requests
from websockets import connect
from websockets.asyncio.client import ClientConnection
from websockets.exceptions import ConnectionClosed

from .types import from_dict, User, Message, RicsInfo

# Exceptions
class AuthorizationFailed(Exception):
    def __init__(self, message: str, json: dict | None = None) -> None:
        super().__init__(message)
        self.json = json

class Disconnected(Exception):
    pass

# Handle state
class ClientState:
    """The current client state. Includes data such as the user list, chat logs, and websocket connection."""
    def __init__(self) -> None:
        self.user_list: list[User]
        self.chat_logs: list[Message]
        self.rics_info: dict[str, str]
        self.socket   : ClientConnection

class Context:
    """An object to store data about the current event context."""
    def __init__(
        self,
        state: ClientState,
        message: typing.Optional[Message] = None,
        user: typing.Optional[User] = None
    ) -> None:
        self.state = state
        self.rics = RicsInfo(
            name = state.rics_info["name"],
            users = state.user_list,
            chat_logs = state.chat_logs
        )
        """The RICS (Realtime Info and Communication System) server we are connected to."""

        if message is not None:
            self.message = message
            """The message we just received. This will be None if this is accessed from anything other then :on_message:."""

        if user is not None:
            self.user = user
            """The user who left or joined the server, or the author of the message we just received."""

    async def send(self, message: str) -> None:
        await self.state.socket.send(orjson.dumps({"type": "message", "data": {"message": message}}), text = True)

    async def reply(self, message: str) -> None:
        await self.send(f"[↑ {self.message.user.name}] {message}")

    def __repr__(self) -> str:
        return f"<Context rics={self.rics} message={getattr(self, 'message', None)} user={getattr(self, 'user', None)}>"

# Main client class
class Client:
    """The main client class, override events on this class and call :run: to start the client."""
    def __init__(self) -> None:
        self.__state = ClientState()
        self.__session = requests.Session()

        # Public attributes (provided just for the hell of it)
        self.user: User | None = None
        """The current user this client is connected as."""
        self.address: str | None = None
        """The address this client is connected to."""

    # Events (for overwriting)
    async def on_connect(self, ctx: Context) -> None:
        """Listen to the :connect: event."""
        pass

    async def on_message(self, ctx: Context) -> None:
        """Listen to the :message: event."""
        pass

    async def on_join(self, ctx: Context) -> None:
        """Listen to the :join: event."""
        pass

    async def on_leave(self, ctx: Context) -> None:
        """Listen to the :leave: event."""
        pass

    # Handle running
    async def __authorize(self, username: str, hex: str, address: str) -> tuple[str, int, str, str]:
        """Given an authorization payload, attempt an authorization request.
        
        Return:
          :host:     (str) -- hostname of the backend
          :port:     (int) -- port of the backend
          :protocol: (str) -- ws(s):// depending on the port
          :auth:     (str) -- authorization code"""
        host, port = (address if ":" in address else f"{address}:443").split(":")
        protocol = "s" if port == "443" else ""

        # Establish authorization
        try:
            response = self.__session.post(
                f"http{protocol}://{host}:{port}/api/join",
                json = {"username": username, "hex": hex, "bot": True},
                timeout = 5
            )
            response.raise_for_status()

            # Handle payload
            payload = response.json()
            if payload["code"] != 200:
                raise AuthorizationFailed("Connection failed!", payload)

            return host, int(port), f"ws{protocol}://", payload["authorization"]

        except requests.RequestException as e:
            raise AuthorizationFailed("Connection failed!", e.response.json() if e.response is not None else None)

    async def __match_event(self, event: dict[str, typing.Any]) -> None:
        match event:
            case {"type": "rics-info", "data": payload}:
                self.__state.chat_logs = [from_dict(Message, message) for message in payload["message-log"]]
                self.__state.user_list = [from_dict(User, user) for user in payload["user-list"]]
                self.__state.rics_info = {"name": payload["name"]}
                await self.on_connect(Context(self.__state))

            case {"type": "message", "data": payload}:
                message = from_dict(Message, payload)

                # Propagate
                await self.on_message(Context(self.__state, message = message, user = message.user))
                self.__state.chat_logs.append(message)

            case {"type": "join", "data": payload}:
                user = from_dict(User, payload["user"])
                if user == self.user:
                    return

                self.__state.user_list.append(user)
                await self.on_join(Context(self.__state, user = user))

            case {"type": "leave", "data": payload}:
                user = from_dict(User, payload["user"])
                self.__state.user_list.remove(user)
                await self.on_leave(Context(self.__state, user = user))

    async def event_loop(self, username: str, hex: str, address: str) -> None:
        """Establish a connection and listen to websocket messages.
        This method shouldn't be called directly, use :Client.run: instead."""

        host, port, protocol, auth = await self.__authorize(username, hex, address)
        self.user, self.address = User(username, hex, False, True), address

        try:
            async with connect(f"{protocol}{host}:{port}/api/ws?authorization={auth}") as socket:
                self.__state.socket = socket
                while socket.state == 1:
                    await self.__match_event(orjson.loads(await socket.recv()))

        except ConnectionClosed:
            raise Disconnected("RICS socket has been disconnected!")

    async def close(self) -> None:
        """Closes the websocket connection."""
        await self.__state.socket.close()

    def run(
        self,
        username: str,
        hex: str,
        address: str
    ):
        """Start the client and run the event loop.

        Arguments:
          :username: (str) -- the username to connect with
          :hex:      (str) -- the hex color code to connect with
          :address:  (str) -- the FQDN to connect to
        """
        asyncio.run(self.event_loop(username, hex, address))



# Copyright (c) 2024 iiPython

# Modules
import sys
from typing import List

from ..vendor.iipython import color, readchar, keys

# Main menu class
class MenuHandler(object):
    def __init__(self) -> None:
        self._config = {
            "colors": {
                "index": "[blue]{}[/]",
                "highlight": "[yellow]{}[/]"
            },
            "suffix": "[dim]<--[/][norm]",
            "indents": {
                "index": "  ",
                "option": "\t"  # \t = 4 spaces
            },
            "enable_index": True
        }

    def config(self, config: dict) -> None:
        for k, v in config.items():
            self._config[k] = v

    def show(self, options: List[str]) -> str:
        iterated, length, index = 0, len(options), 0
        while True:
            sys.stdout.write(f"\033[{length}F" * iterated)

            # Handle control
            if iterated:
                key = readchar()
                if key in [keys.UP, "w"]:
                    index = length - 1 if index - 1 < 0 else index - 1

                elif key in [keys.DOWN, "s"]:
                    index = 0 if index + 1 >= length else index + 1

                elif key == keys.ENTER:
                    sys.stdout.write("\n" * length)
                    sys.stdout.flush()
                    return options[index]

                elif key == keys.CTRL_C:
                    raise KeyboardInterrupt

            # Print out options
            _i = self._config["indents"]
            for i, option in enumerate(options):
                index_text = color(self._config["colors"]["index"].format(i + 1))
                if i == index:
                    option_text = color(self._config["colors"]["highlight"].format(option))
                    suffix_text = color(self._config["suffix"]) if self._config["suffix"] else ""

                else:
                    option_text, suffix_text = option, ""

                index_text = f"{_i['index']}{index_text}" if self._config["enable_index"] else ""
                sys.stdout.write(f"\033[2K{index_text}{_i['option']}{option_text} {suffix_text}\n")

            iterated = 1

menu = MenuHandler()



# Copyright (c) 2024 iiPython

# Modules
import orjson
from nanoid import generate
from types import FunctionType
from websockets import WebSocketCommonProtocol

# Main class
class ORJSONWebSocket():
    def __init__(self, ws: WebSocketCommonProtocol) -> None:
        self.ws = ws
        self.callbacks = {}

    def recv(self) -> dict:
        return orjson.loads(self.ws.recv())

    def send(self, data: dict) -> None:
        self.ws.send(orjson.dumps(data).decode())

    def callback(self, payload: dict, callback: FunctionType) -> None:
        callback_id = generate()
        self.callbacks[callback_id] = callback

        # Send off callback
        if "data" not in payload:
            payload["data"] = {"callback": callback_id}

        else:
            payload["data"]["callback"] = callback_id

        self.send(payload)



# Copyright (c) 2024 iiPython

# Modules
import os
import base64
import binascii

from fastapi import Response
from requests import RequestException
from fastapi.responses import JSONResponse

from nightwatch.rics import app, session
from nightwatch.logging import log

# Exceptions
class IllegalURL(Exception):
    pass

# Handle image forwarding
PROXY_SIZE_LIMIT = 10 * (1024 ** 2)

FORWARD_DOMAIN = os.getenv("DOMAIN") 
if FORWARD_DOMAIN is None:
    log.warn("images", "DOMAIN environment variable not set! Image forwarding unprotected!")

# Routing
@app.get("/api/fwd/{public_url:str}", response_model = None)
async def forward_image(public_url: str) -> Response | JSONResponse:
    try:
        new_url = f"https://{base64.b64decode(public_url.replace('_', '/'), validate = True).decode('ascii').rstrip('/')}"
        if FORWARD_DOMAIN and FORWARD_DOMAIN in new_url:
            raise IllegalURL

    except (binascii.Error, UnicodeDecodeError):
        return JSONResponse({"code": 400, "message": "Failed to contact the specified URI."}, status_code = 400)

    except IllegalURL:
        return JSONResponse({"code": 400, "message": "Requested URL contains an illegal string!"}, status_code = 400)

    try:
        data = b""
        with session.get(new_url, stream = True) as response:
            response.raise_for_status()
            for chunk in response.iter_content(PROXY_SIZE_LIMIT):
                data += chunk
                if len(data) >= PROXY_SIZE_LIMIT:
                    return JSONResponse({"code": 400, "message": "Specified URI contains data above size limit."}, status_code = 400)

            return Response(
                data,
                response.status_code,
                {
                    k: v
                    for k, v in response.headers.items() if k in ["Content-Type", "Cache-Control"]
                }
            )

    except RequestException:
        return JSONResponse({"code": 400, "message": "Failed to contact the specified URI."}, status_code = 400)



__author__ = "iiPython"
__version__ = "1.1.6-3"
__license__ = "MIT"
__copyright__ = "Copyright 2022 iiPython"

from .iikp import readchar, keys
from .term import color, clear, cprint, to_ansi, colormap



# Copyright (c) 2024 iiPython

# Modules
import typing
from dataclasses import dataclass, fields, is_dataclass

# Typing
T = typing.TypeVar("T")

def from_dict(cls: typing.Type[T], data: dict) -> T:
    if not is_dataclass(cls):
        raise ValueError(f"{cls} is not a dataclass")

    field_types = {f.name: f.type for f in fields(cls)}
    instance_data = {}

    for key, value in data.items():
        if key in field_types:
            field_type = field_types[key]
            if is_dataclass(field_type) and isinstance(value, dict):
                instance_data[key] = from_dict(field_type, value)  # type: ignore

            else:
                instance_data[key] = value

    return cls(**instance_data)

@dataclass
class User:
    name: str
    """The name of this user."""
    hex: str
    """The hex color code of this user, without a leading hashtag."""
    admin: bool
    """Status of whether or not this user is an admin."""
    bot: bool
    """Status of whether or not this user is a bot."""

    def __repr__(self) -> str:
        return f"<User name='{self.name}' hex='{self.hex}' admin={self.admin} bot={self.bot}>"

@dataclass
class Message:
    user: User
    """The :User: object who sent this message."""
    message: str
    """The raw text content of this message."""
    time: int
    """The time this message was sent in seconds since the epoch."""

    def __repr__(self) -> str:
        return f"<Message user='{self.user}' message='{self.message}' time={self.time}>"

@dataclass
class RicsInfo:
    name: str
    """The name of the RICS server we are connected to."""
    users: list[User]
    """List of :User: objects that are connected to this server."""
    chat_logs: list[Message]
    """List of :Message: objects consisting of chat logs.
    This will be the last 25 messages sent if you just joined, otherwise it will build over time."""

    def __repr__(self) -> str:
        return f"<RicsInfo name='{self.name}' users=[...] chat_logs=[...]>"



# Copyright (c) 2024 iiPython

# Modules
import shlex
from typing import Tuple
from datetime import datetime
from types import FunctionType

import urwid
import emoji

from .commands import commands
from .wswrap import ORJSONWebSocket
from ..vendor.scroll import Scrollable, ScrollBar

from nightwatch.client import config

# Input edit class
class InputEdit(urwid.Edit):
    def __init__(self, send_message: FunctionType) -> None:
        self.send_message = send_message
        super().__init__("> ", wrap = "clip")

    def keypress(self, size: Tuple[int], key: str) -> None:
        if key != "enter":
            return super().keypress(size, key)

        elif not self.edit_text.strip():
            return

        self.send_message(self.edit_text)
        self.edit_text = ""

# Main UI class
class NightwatchUI():
    def __init__(self, websocket: ORJSONWebSocket) -> None:
        self.last_author, self.last_time = None, 0
        self.websocket = websocket

        # Main urwid setup
        self.pile = urwid.Pile([])
        self.scroll = Scrollable(urwid.Filler(self.pile, valign = "top"))
        self.frame = urwid.Frame(ScrollBar(self.scroll), footer = InputEdit(self.send_message), focus_part = "footer")

        # Initialize commands
        self.commands = {}
        for command in commands:
            command = command(self, self.add_message)
            self.commands[command.name] = command

    def length(self, text: str) -> int:
        return int(len(text.encode(encoding = "utf_16_le")) / 2)

    def send_message(self, text: str) -> None:
        if text[0] == "/":
            try:
                splices = shlex.split(text.replace("\\", "\\\\"))
                if splices[0][1:] in self.commands:
                    response = self.commands[splices[0][1:]].on_execute(splices[1:])
                    if response is None:
                        return  # Internal command, do not send in chat lmao

                    text = response

            except ValueError:
                pass

        self.websocket.send({"type": "message", "data": {"message": text}})

    def construct_message(self, author: str, content: str, user_color: str = "gray") -> None:
        visible_author = author if author != self.last_author else " " * self.length(author)
        now, time_string = datetime.now(), ""
        if (author != self.last_author) or ((now - self.last_time).total_seconds() > 300):
            time_string = now.strftime("%I:%M %p" if config["client.time_format"] != "24h" else "%H:%M") + "  "  # Right padding for the scrollbar

        self.pile.contents.append((urwid.Columns([
            (self.length(visible_author), urwid.Text((user_color, visible_author))),
            (3, urwid.Text(("sep", " | "))),
            ("weight", 4, urwid.Text(content)),
            (len(time_string) + 2, urwid.Text(("time", time_string), align = "right"))  # +2 adds left padding
        ]), self.pile.options()))
        self.last_author, self.last_time = author, now

    def add_message(self, author: str, content: str, user_color: str = "gray") -> None:
        self.construct_message(author, emoji.emojize(content), user_color)

        # Scroll to the end (if we haven't manually moved)
        if len(self.pile.contents) > 50:
            self.pile.contents = self.pile.contents[1:]

        self.scroll.to_end()
        self.loop.screen.clear()
        self.loop.draw_screen()

    def on_message(self, data: dict) -> None:
        if not hasattr(self, "loop"):
            return  # We aren't even initialized yet

        match data:
            case {"type": "response", "data": {"callback": callback}}:
                callbackfn = self.websocket.callbacks[callback]
                del self.websocket.callbacks[callback]
                return callbackfn(data)

            case {"type": "message", "data": {"user": user, "message": message}}:
                color_code = f"user-{user['name']}"
                self.loop.screen.register_palette_entry(color_code, "yellow", "", foreground_high = f"#{user['hex']}")

                # Push message to screen
                self.add_message(user["name"], message, color_code)

            case _:
                exit(f"Nightwatch Exception\n{'=' * 50}\n\n{data['data']['message']}")

    def on_ready(self, loop: urwid.MainLoop, payload: dict) -> None:
        self.loop = loop
        self.construct_message("Nightwatch", f"You have connected to {payload['name']}.")



# Keycodes
UP         = 72
DOWN       = 80
LEFT       = 75
RIGHT      = 77
ENTER      = 13
SPACE      = 32

INSERT     = 82
DELETE     = 83
HOME       = 71
END        = 79
PG_UP      = 73
PG_DOWN    = 81

ESCAPE     = 27
BACKSPACE  = 8

CTRL_C     = 3

# Function keys
F1         = 59
F2         = 60
F3         = 61
F4         = 62
F5         = 63
F6         = 64
F7         = 65
F8         = 66
F9         = 67
F10        = 68
F11        = 133
F12        = 134

# Key aliases
ESC        = ESCAPE
PAGE_DN    = PG_DOWN
PAGE_UP    = PG_UP
INS        = INSERT
DEL        = DELETE

# Keymap
MAP = [globals()[d] for d in globals() if not d.startswith("__") and d != "MAP" and isinstance(globals()[d], int)]
for _ in MAP:
    while MAP.count(_) > 1:
        MAP.remove(_)



# Copyright (c) 2024 iiPython

# Modules
import typing
import shutil
import mimetypes
from pathlib import Path

from nanoid import generate
from fastapi import UploadFile
from fastapi.responses import JSONResponse, FileResponse
from pydantic import BaseModel, Field, field_validator

from nightwatch.rics import app, config

# Initialization
SIZE_LIMIT = 400 * (1024 ** 2)  # 100MB
CHUNK_LIMIT = SIZE_LIMIT / 4

UPLOAD_LOCATION = Path(config["file_upload_location"] or (config.config_path.parent / "file_uploads"))
UPLOAD_LOCATION.mkdir(parents = True, exist_ok = True)

app.state.uploads = {}

# Models
class FileCreationModel(BaseModel):
    size: typing.Annotated[int, Field(ge = 1, le = SIZE_LIMIT)]  # 1B - SIZE_LIMIT
    name: str

    # Ensure that our filename conforms to ext4 storage requirements
    @field_validator("name")
    def validate_filename(cls, value: str) -> str:
        if "/" in value or "\0" in value:
            raise ValueError("Filename contains / or a null character!")

        if len(value.encode("utf-8")) > 255:
            raise ValueError("Filename must be <= 255 bytes in length!")

        return value

# Handle routing
@app.post("/api/file")
async def route_file_create(file: FileCreationModel) -> JSONResponse:
    file_id = generate()

    # Save this upload
    app.state.uploads[file_id] = file.model_dump()
    return JSONResponse({
        "code": 200,
        "data": {
            "file_id": file_id
        }
    })

@app.post("/api/file/{file_id:str}")
async def route_file_upload(upload: UploadFile, file_id: str) -> JSONResponse:
    if file_id not in app.state.uploads:
        return JSONResponse({"code": 403}, status_code = 403)

    target = app.state.uploads[file_id]
    destination = UPLOAD_LOCATION / file_id / target["name"]

    if not destination.parent.is_dir():
        destination.parent.mkdir()

    existing_size = target.get("written_bytes", 0)
    if existing_size > target["size"]:
        shutil.rmtree(destination.parent)
        del app.state.uploads[file_id]
        return JSONResponse({"code": 400, "message": "File exceeds size limit."}, status_code = 400)

    # Check filesize of this chunk
    upload.file.seek(0, 2)  # Go to end of file
    chunk_size = upload.file.tell()

    if chunk_size > CHUNK_LIMIT:
        if destination.is_file():
            shutil.rmtree(destination.parent)

        del app.state.uploads[file_id]
        return JSONResponse({"code": 400, "message": "Chunk exceeds size limit."}, status_code = 400)

    if existing_size + chunk_size > target["size"]:
        if destination.is_file():
            shutil.rmtree(destination.parent)

        del app.state.uploads[file_id]
        return JSONResponse({"code": 400, "message": "File exceeds size limit."}, status_code = 400)

    # Save to disk
    app.state.uploads[file_id]["written_bytes"] = existing_size + chunk_size
    with destination.open("ab") as handle:
        upload.file.seek(0)
        handle.write(await upload.read())

    return JSONResponse({"code": 200, "data": {"current_size": target["written_bytes"]}})

@app.post("/api/file/{file_id:str}/finalize")
async def route_file_finalize(file_id: str) -> JSONResponse:
    if file_id not in app.state.uploads:
        return JSONResponse({"code": 403}, status_code = 403)

    target = app.state.uploads[file_id]
    if target.get("written_bytes", 0) < 1:
        if (UPLOAD_LOCATION / file_id).is_dir():
            shutil.rmtree(UPLOAD_LOCATION / file_id)

        del app.state.uploads[file_id]
        return JSONResponse({"code": 400, "message": "No data has been written to file."}, status_code = 400)

    del app.state.uploads[file_id]
    return JSONResponse({"code": 200, "data": {"path": f"{file_id}/{target['name']}"}})

@app.get("/api/file/{file_id:str}/info")
async def route_file_info(file_id: str) -> JSONResponse:
    file_path = UPLOAD_LOCATION / file_id
    if not (file_path.is_dir() and file_path.relative_to(UPLOAD_LOCATION)):
        return JSONResponse({"code": 404}, status_code = 404)

    actual_file = next(file_path.iterdir())
    return JSONResponse({
        "code": 200,
        "data": {
            "size": actual_file.stat().st_size,
            "name": actual_file.name
        }
    })

@app.get("/file/{file_id:str}/{file_name:str}", response_model = None)
async def route_file_download(file_id: str, file_name: str) -> FileResponse | JSONResponse:
    file_path = UPLOAD_LOCATION / file_id / file_name
    if not (file_path.is_file() and file_path.relative_to(UPLOAD_LOCATION)):
        return JSONResponse({"code": 404}, status_code = 404)

    content_type = mimetypes.guess_type(file_path.name)[0]
    return FileResponse(
        file_path,
        headers = {
            "Content-Type": content_type or "application/octet-stream",
            "Content-Disposition": "inline" if content_type else "attachment"
        }
    )



# Copyright 2022 iiPython

# Modules
import sys
import tty
import termios
from . import keys

# Readchar function
def readchar():
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)

        if ch == "\x1b":
            ch = sys.stdin.read(2)
            if ch == "[D":
                return keys.LEFT

            elif ch == "[C":
                return keys.RIGHT

            elif ch == "[A":
                return keys.UP

            elif ch == "[B":
                return keys.DOWN

        elif ch == "\r":
            return keys.ENTER

        elif ch == "\x7f":
            return keys.BACKSPACE

        elif ch == "\x03":
            return keys.CTRL_C

    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return ch



# Copyright (c) 2024 iiPython

# Modules
import os
import json
import typing
from pathlib import Path

# Initialization
data_override = os.getenv("DATA_OVERRIDE")
if data_override is None:
    config_path = Path.home() / ".local/share/nightwatch"
    if os.name == "nt":
        appdata = os.getenv("APPDATA")
        if appdata is None:
            exit(r"Nightwatch: %APPDATA% is None, no place available to store data at.")

        config_path = Path(appdata) / "Nightwatch"

else:
    config_path = Path(data_override)

config_path.mkdir(exist_ok = True, parents = True)

# Configuration class
class Configuration():
    def __init__(self, config_path: Path) -> None:
        self.config, self.config_path = {}, config_path
        if config_path.is_file():
            self.config = json.loads(self.config_path.read_text())

    def __getitem__(self, item: str) -> typing.Any:
        v = self.config
        for k in item.split("."):
            if k not in v:
                return None

            v = v[k]

        return v

    def set(self, key: str, value: typing.Any) -> None:
        v = self.config
        for k in key.split(".")[:-1]:
            if k not in v:
                v[k] = {}

            v = v[k]

        v[key.split(".")[-1]] = value
        self.config_path.write_text(json.dumps(self.config, indent = 4))

    def reset(self) -> None:
        self.config = {}
        self.config_path.write_text(json.dumps({}))

# Modern configuration system
def fetch_config(config_module: str) -> Configuration:
    return Configuration(config_path / f"{config_module}.json")
